# -*- mode: python ; coding: utf-8 -*-
import json
import pytest

try:
    import eth_account
except ImportError:
    eth_account			= None
try:
    from Crypto.Cipher	import AES
    from Crypto.Protocol.KDF import scrypt
except ImportError:
    AES				= None
    scrypt			= None

import shamir_mnemonic

from .			import account, create, addresses, addressgroups, accountgroups, Account
from .recovery		import recover

from .dependency_test	import substitute, nonrandom_bytes, SEED_XMAS


def test_account():
    acct			= account( SEED_XMAS )
    assert acct.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert acct.path == "m/44'/60'/0'/0/0"
    assert acct.key == '178870009416174c9697777b1d94229504e83f25b1605e7bb132aa5b88da64b6'

    acct			= account( SEED_XMAS, path="m/44'/60'/0'/0/1" )
    assert acct.address == '0x3b774e485fC818F0f377FBA657dfbF92B46f8504'
    assert acct.path == "m/44'/60'/0'/0/1"

    acct			= account( SEED_XMAS, crypto='Bitcoin' )
    assert acct.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'
    assert acct.path == "m/84'/0'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Bitcoin', format='Legacy' )
    assert acct.address == '19FQ983heQEBXmopVNyJKf93XG7pN7sNFa'
    assert acct.path == "m/44'/0'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Bitcoin', format='SegWit' )
    assert acct.address == '3HxUpD7E8Y31vDDgDq1VFdNXWViAgBjYJe'
    assert acct.path == "m/44'/0'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Litecoin' )
    assert acct.address == 'ltc1qfjepkelqd3jx4e73s7p79lls6kqvvmak5pxy97'
    assert acct.path == "m/84'/2'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Litecoin', format='Legacy' )
    assert acct.address == 'LeyK1dbc5qKdKC9TvkygMTeoHixR3z1XG3'
    assert acct.path == "m/44'/2'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Litecoin', format='SegWit' )
    assert acct.address == 'MPULjvY9dNjpNkgbwhfJtD7N6Lbfam1XsP'
    assert acct.path == "m/44'/2'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Dogecoin' )
    assert acct.address == 'DQCnF49GwQ5auL3u5c2uow62XS57RCA2r1'
    assert acct.path == "m/44'/3'/0'/0/0"


@pytest.mark.skipif( not scrypt or not eth_account,
                     reason="pip install slip39[wallet] to support private key encryption" )
def test_account_encrypt():
    """Ensure BIP-38 and Ethereum JSON wallet encryption and recovery works."""

    acct			= account( SEED_XMAS, crypto='Bitcoin' )
    assert acct.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'
    assert acct.crypto == 'BTC'
    assert acct.path == "m/84'/0'/0'/0/0"
    assert acct.legacy_address() == "134t1ktyF6e4fNrJR8L6nXtaTENJx9oGcF"

    bip38_encrypted		= acct.encrypted( 'password' )
    assert bip38_encrypted == '6PYKmUhfJa5m1NR2zUaeHC3wUzGDmb1seSEgQHK7PK5HaVRHQSp7N4ytVf'

    acct_reco			= Account( crypto='Bitcoin' ).from_encrypted( bip38_encrypted, 'password' )
    assert acct_reco.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'
    assert acct.crypto == 'BTC'
    assert acct.path == "m/84'/0'/0'/0/0"  # The default; assumed...
    assert acct_reco.legacy_address() == "134t1ktyF6e4fNrJR8L6nXtaTENJx9oGcF"

    acct			= account( SEED_XMAS, crypto='Ethereum' )
    assert acct.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert acct.crypto == 'ETH'
    assert acct.path == "m/44'/60'/0'/0/0"

    json_encrypted		= acct.encrypted( 'password' )
    assert json.loads( json_encrypted ).get( 'address' ) == '336cbeab83accdb2541e43d514b62dc6c53675f4'

    acct_reco			= Account( crypto='ETH' ).from_encrypted( json_encrypted, 'password' )
    assert acct_reco.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert acct.crypto == 'ETH'

    # Some test cases from https://en.bitcoin.it/wiki/BIP_0038

    # No compression, no EC multiply.  These tests produce the correct private key hex, but the
    # address hash verification check fails.  I suspect that they were actually created with a
    # *different* passphrase...
    assert Account( crypto='BTC' ).from_encrypted(
        '6PRVWUbkzzsbcVac2qwfssoUJAN1Xhrg6bNk8J7Nzm5H7kxEbn2Nh2ZoGg',
        'TestingOneTwoThree',
        strict=False,
    ).key.upper() == 'CBF4B9F70470856BB4F40F80B87EDB90865997FFEE6DF315AB166D713AF433A5'

    assert Account( crypto='BTC' ).from_encrypted(
        '6PRNFFkZc2NZ6dJqFfhRoFNMR9Lnyj7dYGrzdgXXVMXcxoKTePPX1dWByq',
        'Satoshi',
        strict=False,
    ).key.upper() == '09C2686880095B1A4C249EE3AC4EEA8A014F11E6F986D0B5025AC1F39AFBD9AE'

    # This weird UTF-8 test I cannot get to pass, regardless of what format I supply the passphrase in..

    # acct_reco			= Account( crypto='BTC' ).from_encrypted(
    #     '6PRW5o9FLp4gJDDVqJQKJFTpMvdsSGJxMYHtHaQBF3ooa8mwD69bapcDQn',
    #      bytes.fromhex('cf9300f0909080f09f92a9'), # '\u03D2\u0301\u0000\U00010400\U0001F4A9'
    # )
    # assert acct_reco.legacy_address() == '16ktGzmfrurhbhi6JGqsMWf7TyqK9HNAeF'

    # No compression, no EC multiply.  These test pass without relaxing the address hash verification check.
    assert Account( crypto='BTC' ).from_encrypted(
        '6PYNKZ1EAgYgmQfmNVamxyXVWHzK5s6DGhwP4J5o44cvXdoY7sRzhtpUeo',
        'TestingOneTwoThree'
    ).key.upper() == 'CBF4B9F70470856BB4F40F80B87EDB90865997FFEE6DF315AB166D713AF433A5'

    assert Account( crypto='BTC' ).from_encrypted(
        '6PYLtMnXvfG3oJde97zRyLYFZCYizPU5T3LwgdYJz1fRhh16bU7u6PPmY7',
        'Satoshi'
    ).key.upper() == '09C2686880095B1A4C249EE3AC4EEA8A014F11E6F986D0B5025AC1F39AFBD9AE'


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_create():
    details			= create(
        "SLIP39 Wallet: Test",
        1,
        dict( fren = (3,5) ),
        SEED_XMAS
    )

    assert details.groups == {
        "fren": ( 3, [
            "academic acid academic acne academic academic academic academic academic academic academic academic academic academic academic academic academic carpet making building",
            "academic acid academic agree depart dance galaxy acrobat mayor disaster quick justice ordinary agency plunge should pupal emphasis security obtain",
            "academic acid academic amazing crush royal faint spit briefing craft floral negative work depend prune adapt merit romp home elevator",
            "academic acid academic arcade cargo unfold aunt spider muscle bedroom triumph theory gather dilemma building similar chemical object cinema salon",
            "academic acid academic axle crush swing purple violence teacher curly total equation clock mailman display husband tendency smug laundry disaster"
        ] ),
    }

    assert len(details.accounts) == 1
    [(eth,btc)] = details.accounts  # The default accounts created are ETH, BTC
    assert eth.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert btc.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'

    assert recover( details.groups['fren'][1][:3] ) == SEED_XMAS


def test_addresses():
    master_secret		= b'\xFF' * 16
    addrs			= list( addresses(
        master_secret	= master_secret,
        crypto		= 'ETH',
        paths		= ".../-9",
    ))
    # print( json.dumps( addrs, indent=4, default=str ))
    assert addrs == [
        (
            "ETH",
            "m/44'/60'/0'/0/0",
            "0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/1",
            "0x8D342083549C635C0494d3c77567860ee7456963"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/2",
            "0x52787E24965E1aBd691df77827A3CfA90f0166AA"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/3",
            "0xc2442382Ae70c77d6B6840EC6637dB2422E1D44e"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/4",
            "0x42a910D380dE132B5227e3277Cc70C3C76a884aC"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/5",
            "0x1A3db5E0422c78F43a35686f0307Da8f22344dE0"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/6",
            "0x19031c515C5d91DB7988D89AAA6F71a5825f5245"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/7",
            "0xaE693156ac600f5B0D58e5090ecf0A578c5Cc0a8"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/8",
            "0x4347541fa648BCE62543a8AbC2901E08017f6A6a"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/9",
            "0xC11235559Dd4c5224a19396C3A14526E92ebba35"
        )
    ]
    addrs			= list( addresses(
        master_secret	= master_secret,
        crypto		= 'BTC',
        paths		= '.../0/-2',
        format		= 'Bech32',
    ))
    print( json.dumps( addrs, indent=4, default=str ))
    assert addrs == [
        (
            "BTC",
            "m/84'/0'/0'/0/0",
            "bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl"
        ),
        (
            "BTC",
            "m/84'/0'/0'/0/1",
            "bc1qnec684yvuhfrmy3q856gydllsc54p2tx9w955c"
        ),
        (
            "BTC",
            "m/84'/0'/0'/0/2",
            "bc1q2snj0zcg23dvjpw7m9lxtu0ap0hfl5tlddq07j"
        ),
    ]


def test_addressgroups():
    master_secret		= b'\xFF' * 16
    addrgrps			= list( enumerate( addressgroups(
        master_secret	= master_secret,
        cryptopaths	= [
            ('ETH', ".../-3"),
            ('BTC', ".../-3"),
            ('LTC', ".../-3"),
            ('Doge', ".../-3"),
        ],
    )))
    # print( addrgrps )
    assert addrgrps == [
        (0, (("ETH", "m/44'/60'/0'/0/0", "0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1"),
             ("BTC", "m/84'/0'/0'/0/0", "bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl"),
             ('LTC', "m/84'/2'/0'/0/0", 'ltc1qe5m2mst9kjcqtfpapaanaty40qe8xtusmq4ake'),
             ('DOGE', "m/44'/3'/0'/0/0", 'DN8PNN3dipSJpLmyxtGe4EJH38EhqF8Sfy'))),
        (1, (("ETH", "m/44'/60'/0'/0/1", "0x8D342083549C635C0494d3c77567860ee7456963"),
             ("BTC", "m/84'/0'/0'/0/1", "bc1qnec684yvuhfrmy3q856gydllsc54p2tx9w955c"),
             ('LTC', "m/84'/2'/0'/0/1", 'ltc1qm0hwvvk28wlyfu3sed66e9yyvmwm35xtfexva3'),
             ('DOGE', "m/44'/3'/0'/0/1",'DJYE9WWaCA1CbV9x23qkcgNX7Yr9YjCebA'))),
        (2, (("ETH", "m/44'/60'/0'/0/2", "0x52787E24965E1aBd691df77827A3CfA90f0166AA"),
             ("BTC", "m/84'/0'/0'/0/2", "bc1q2snj0zcg23dvjpw7m9lxtu0ap0hfl5tlddq07j"),
             ('LTC', "m/84'/2'/0'/0/2", 'ltc1qx3r3efsmupn34gmwu25fu39tn4h79cjfwvlpfu'),
             ('DOGE', "m/44'/3'/0'/0/2",'DQfJcJzLFW9LJPJXNkLeq1WqPfLsRq47Jj'))),
        (3, (("ETH", "m/44'/60'/0'/0/3", "0xc2442382Ae70c77d6B6840EC6637dB2422E1D44e"),
             ("BTC", "m/84'/0'/0'/0/3", "bc1qxwekjd46aa5n0s3dtsynvtsjwsne7c5f5w5dsd"),
             ('LTC', "m/84'/2'/0'/0/3", 'ltc1qnqzyear8kct0yjzupe2pxtq0mwee5kl642mj78'),
             ('DOGE', "m/44'/3'/0'/0/3", 'DLVPiM5763cyNJfoa13cv4kV3b87FgVMCS'))),
    ]


def test_accountgroups():
    master_secret		= b'\xFF' * 16
    acctgrps			= list( accountgroups(
        master_secret	= master_secret,
        cryptopaths	= [
            ('ETH', ".../-3"),
            ('BTC', ".../-3"),
        ],
    ))
    # print( json.dumps( acctgrps, default=repr ))
    assert len(acctgrps) == 4



import base58
import codecs
import hashlib
import itertools
import json
import logging
import math
import re
import secrets

from collections	import namedtuple
from typing		import Dict, List, Sequence, Tuple, Union, Callable

from shamir_mnemonic	import generate_mnemonics

import hdwallet

# Support for private key encryption via BIP-38 and Ethereum JSON wallet is optional; pip install slip39[wallet]
try:
    from Crypto.Cipher	import AES
    from Crypto.Protocol.KDF import scrypt
except ImportError:
    AES				= None
    scrypt			= None
try:
    import eth_account
except ImportError:
    eth_account			= None

from .defaults		import BITS_DEFAULT, BITS, MNEM_ROWS_COLS, GROUP_REQUIRED_RATIO, CRYPTO_PATHS
from .util		import ordinal


RANDOM_BYTES			= secrets.token_bytes

log				= logging.getLogger( __package__ )


def paper_wallet_available():
    """Determine if encrypted BIP-38 and Ethereum JSON Paper Wallets are available."""
    return AES and scrypt and eth_account


def path_edit(
    path: str,
    edit: str,
):
    """Replace the current path w/ the new path, either entirely, or if only partially if a continuation
    '../' followed by some new path segments is provided.

    """
    if edit.startswith( '.' ):
        new_segs	= edit.lstrip( './' ).split( '/' )
        cur_segs	= path.split( '/' )
        log.info( f"Using {edit} to replace last {len(new_segs)} of {path} with {'/'.join(new_segs)}" )
        if len( new_segs ) >= len( cur_segs ):
            raise ValueError( f"Cannot use {edit} to replace last {len(new_segs)} of {path} with {'/'.join(new_segs)}" )
        res_segs	= cur_segs[:len(cur_segs)-len(new_segs)] + new_segs
        return '/'.join( res_segs )
    else:
        return edit


class Account( hdwallet.HDWallet ):

    """Supports producing Legacy addresses for Bitcoin, and Litecoin.  Doge (D...) and Ethereum (0x...)
    addresses use standard BIP44 derivation.


    | Crypto | Semantic | Path             | Address |
    |        |          |                  | <       |
    |--------+----------+------------------+---------|
    | ETH    | Legacy   | m/44'/60'/0'/0/0 | 0x...   |
    | BTC    | Legacy   | m/44'/ 0'/0'/0/0 | 1...    |
    |        | SegWit   | m/44'/ 0'/0'/0/0 | 3...    |
    |        | Bech32   | m/84'/ 0'/0'/0/0 | bc1...  |
    | LTC    | Legacy   | m/44'/ 2'/0'/0/0 | L...    |
    |        | SegWit   | m/44'/ 2'/0'/0/0 | M...    |
    |        | Bech32   | m/84'/ 2'/0'/0/0 | ltc1... |
    | DOGE   | Legacy   | m/44'/ 3'/0'/0/0 | D...    |

    """
    CRYPTO_NAMES		= dict(				# Currently supported
        ethereum	= 'ETH',
        bitcoin		= 'BTC',
        litecoin	= 'LTC',
        dogecoin	= 'DOGE',
    )
    CRYPTOCURRENCIES		= set( CRYPTO_NAMES.values() )

    ETHJS_ENCRYPT		= set( ('ETH',) )			# Can be encrypted w/ Ethereum JSON wallet
    BIP38_ENCRYPT		= CRYPTOCURRENCIES - ETHJS_ENCRYPT      # Can be encrypted w/ BIP-38

    CRYPTO_FORMAT		= dict(
        ETH		= "legacy",
        BTC		= "bech32",
        LTC		= "bech32",
        DOGE		= "legacy",
    )
    FORMATS		= ("legacy", "segwit", "bech32")

    CRYPTO_FORMAT_PATH		= dict(
        ETH		= dict(
            legacy	= "m/44'/60'/0'/0/0",
        ),
        BTC		= dict(
            legacy	= "m/44'/0'/0'/0/0",
            segwit	= "m/44'/0'/0'/0/0",
            bech32	= "m/84'/0'/0'/0/0",
        ),
        LTC		= dict(
            legacy	= "m/44'/2'/0'/0/0",
            segwit	= "m/44'/2'/0'/0/0",
            bech32	= "m/84'/2'/0'/0/0",
        ),
        DOGE		= dict(
            legacy	="m/44'/3'/0'/0/0",
        )
    )

    @classmethod
    def path_default( cls, crypto, format=None ):
        """Return the default derivation path for the given crypto, based on its currently selected default
        address format.

        """
        crypto			= cls.supported( crypto )
        format			= format.lower() if format else cls.address_format( crypto )
        if format not in cls.CRYPTO_FORMAT_PATH[crypto]:
            raise ValueError( f"{format} not supported for {crypto}; specify one of {', '.join( cls.CRYPTO_FORMAT_PATH[crypto].keys() )}" )
        return cls.CRYPTO_FORMAT_PATH[crypto][format]

    @classmethod
    def address_format( cls, crypto, format=None ):
        """Get or set the desired default address format for the specified supported crypto.  Future
        instances of Address created for the crypto will use the specified address format and its
        default derivation path.

        """
        crypto			= cls.supported( crypto )
        if format is None:
            return cls.CRYPTO_FORMAT[crypto]

        format			= format.lower() if format else None
        if format not in cls.FORMATS:
            raise ValueError( f"{crypto} address format {format!r} not recognized; specify one of {', '.join( cls.FORMATS )}" )
        cls.CRYPTO_FORMAT[crypto]	= format

    @classmethod
    def supported( cls, crypto ):
        """Validates that the specified cryptocurrency is supported and returns the normalized short name
        for it, or raises an a ValueError.  Eg. "Ethereum" --> "ETH"

        """
        validated			= cls.CRYPTO_NAMES.get(
            crypto.lower(),
            crypto.upper() if crypto.upper() in cls.CRYPTOCURRENCIES else None
        )
        if validated:
            return validated
        raise ValueError( f"{crypto} not presently supported; specify {', '.join( cls.CRYPTOCURRENCIES )}" )

    def __init__( self, crypto, format=None ):
        crypto			= Account.supported( crypto )
        self.format		= format.lower() if format else Account.address_format( crypto )
        if self.format in ("legacy", "segwit",):
            self.hdwallet	= hdwallet.BIP44HDWallet( symbol=crypto )
        elif self.format in ("bech32",):
            self.hdwallet	= hdwallet.BIP84HDWallet( symbol=crypto )
        else:
            raise ValueError( f"{crypto} does not support address format {self.format}" )

    def from_seed( self, seed: str, path: str = None ) -> "Account":
        """Derive the Account from the supplied seed and (optionally) path; uses the default derivation path
        for the Account address format, if None provided.

        """
        if type( seed ) is bytes:
            seed		= codecs.encode( seed, 'hex_codec' ).decode( 'ascii' )
        self.hdwallet.from_seed( seed )
        self.from_path( path )
        return self

    def from_path( self, path: str = None ) -> "Account":
        """Change the Account to derive from the provided path.

        If a partial path is provided (eg "...1'/0/3"), then use it to replace the given segments in
        current account path, leaving the remainder alone.

        """
        if path:
            path		= path_edit( self.path, path )
        else:
            path		= Account.path_default( self.crypto, self.format )
        self.hdwallet.clean_derivation()
        self.hdwallet.from_path( path )
        return self

    @property
    def address( self ):
        if self.format == "legacy":
            return self.legacy_address()
        elif self.format == "segwit":
            return self.segwit_address()
        elif self.format == "bech32":
            return self.bech32_address()
        raise ValueError( f"Unknown addresses semantic: {self.format}" )

    def legacy_address( self ):
        return self.hdwallet.p2pkh_address()

    def segwit_address( self ):
        return self.hdwallet.p2sh_address()

    def bech32_address( self ):
        return self.hdwallet.p2wpkh_address()

    @property
    def crypto( self ):
        return self.hdwallet._cryptocurrency.SYMBOL

    @property
    def path( self ):
        return self.hdwallet.path()

    @property
    def key( self ):
        return self.hdwallet.private_key()

    def from_private_key( self, private_key ):
        self.hdwallet.from_private_key( private_key )
        return self

    def encrypted( self, passphrase ):
        """Output the appropriately encrypted private key for this cryptocurrency.  Ethereum uses
        encrypted JSON wallet standard, Bitcoin et.al. use BIP-38 encrypted private keys."""
        if self.crypto in self.ETHJS_ENCRYPT:
            if not eth_account:
                raise NotImplementedError( "The eth-account module is required to support Ethereum JSON wallet encryption; pip install slip39[wallet]" )
            wallet_dict		= eth_account.Account.encrypt( self.key, passphrase )
            return json.dumps( wallet_dict, separators=(',',':') )
        return self.bip38( passphrase )

    def from_encrypted( self, encrypted_privkey, passphrase, strict=True ):
        """Import the appropriately decrypted private key for this cryptocurrency."""
        if self.crypto in self.ETHJS_ENCRYPT:
            if not eth_account:
                raise NotImplementedError( "The eth-account module is required to support Ethereum JSON wallet decryption; pip install slip39[wallet]" )
            private_hex		= bytes( eth_account.Account.decrypt( encrypted_privkey, passphrase )).hex()
            self.from_private_key( private_hex )
            return self
        return self.from_bip38( encrypted_privkey, passphrase=passphrase, strict=strict )

    def bip38( self, passphrase, flagbyte=b'\xe0' ):
        """BIP-38 encrypt the private key"""
        if not scrypt or not AES:
            raise NotImplementedError( "The scrypt module is required to support BIP-38 encryption; pip install slip39[wallet]" )
        if self.crypto not in self.BIP38_ENCRYPT:
            raise NotImplementedError( f"{self.crypto} does not support BIP-38 private key encryption" )
        private_hex		= self.key
        addr			= self.legacy_address().encode( 'UTF-8' )  # Eg. b"184xW5g..."
        ahash			= hashlib.sha256( hashlib.sha256( addr ).digest() ).digest()[0:4]
        if isinstance( passphrase, str ):
            passphrase		= passphrase.encode( 'UTF-8' )
        key			= scrypt( passphrase, salt=ahash, key_len=64, N=16384, r=8, p=8 )
        derivedhalf1		= key[0:32]
        derivedhalf2		= key[32:64]
        aes			= AES.new( derivedhalf2, AES.MODE_ECB )
        enchalf1		= aes.encrypt( ( int( private_hex[ 0:32], 16 ) ^ int.from_bytes( derivedhalf1[ 0:16], 'big' )).to_bytes( 16, 'big' ))
        enchalf2		= aes.encrypt( ( int( private_hex[32:64], 16 ) ^ int.from_bytes( derivedhalf1[16:32], 'big' )).to_bytes( 16, 'big' ))
        prefix			= b'\x01\x42'
        encrypted_privkey	= prefix + flagbyte + ahash + enchalf1 + enchalf2
        # Encode the encrypted private key to base58, adding the 4-byte base58 check suffix
        return base58.b58encode_check( encrypted_privkey ).decode( 'UTF-8' )

    def from_bip38( self, encrypted_privkey, passphrase, strict=True ):
        """Bip-38 decrypt and import the private key."""
        if not scrypt or not AES:
            raise NotImplementedError( "The scrypt module is required to support BIP-38 decryption; pip install slip39[wallet]" )
        if self.crypto not in self.BIP38_ENCRYPT:
            raise NotImplementedError( f"{self.crypto} does not support BIP-38 private key decryption" )
        # Decode the encrypted private key from base58, discarding the 4-byte base58 check suffix
        d			= base58.b58decode_check( encrypted_privkey )
        assert len( d ) == 43 - 4, \
            f"BIP-38 encrypted key should be 43 bytes long, not {len( d ) + 4} bytes"
        pre,flag,ahash,eh1,eh2	= d[0:2],d[2:3],d[3:7],d[7:7+16],d[7+16:7+32]
        assert pre == b'\x01\x42', \
            f"Unrecognized BIP-38 encryption prefix: {pre!r}"
        assert flag in ( b'\xc0', b'\xe0' ), \
            f"Unrecognized BIP-38 flagbyte: {flag!r}"
        if isinstance( passphrase, str ):
            passphrase		= passphrase.encode( 'UTF-8' )
        key			= scrypt( passphrase, salt=ahash, key_len=64, N=16384, r=8, p=8 )
        derivedhalf1		= key[0:32]
        derivedhalf2		= key[32:64]
        aes			= AES.new( derivedhalf2, AES.MODE_ECB )
        dechalf2		= aes.decrypt( eh2 )
        dechalf1		= aes.decrypt( eh1 )
        priv			= dechalf1 + dechalf2
        priv			= ( int.from_bytes( priv, 'big' ) ^ int.from_bytes( derivedhalf1, 'big' )).to_bytes( 32, 'big' )
        # OK, we have the Private Key; we can recover the account, then verify the
        # remainder of the checks
        private_hex		= codecs.encode( priv, 'hex_codec' ).decode( 'ascii' )
        self.from_private_key( private_hex )
        addr			= self.legacy_address().encode( 'UTF-8' )  # Eg. b"184xW5g..."
        ahash_confirm		= hashlib.sha256( hashlib.sha256( addr ).digest() ).digest()[0:4]
        if ahash_confirm != ahash:
            warning		= f"BIP-38 address hash verification failed ({ahash_confirm.hex()} != {ahash.hex()}); password may be incorrect."
            if strict:
                raise AssertionError( warning )
            else:
                log.warning( warning )
        return self


def path_parser(
    paths: str,
    allow_unbounded: bool	= True,
) -> Tuple[str, Dict[str, Callable[[], int]]]:
    """Create a format and a dictionary of iterators to feed into it.

    Supports paths with an arbitrary prefix, eg. 'm/' or '../'
    """
    path_segs			= paths.split( '/' )
    unbounded			= False
    ranges			= {}

    for i,s in list( enumerate( path_segs )):
        if '-' not in s:
            continue
        c			= chr(ord('a')+i)
        tic			= s.endswith( "'" )
        if tic:
            s			= s[:-1]
        b,e			= s.split( '-' )
        b			= int( b or 0 )
        if e:
            e			= int( e )
            ranges[c]		= lambda b=b,e=e: range( b, e+1 )
        else:
            assert allow_unbounded and not ( unbounded or ranges ), \
                f"{'Only first' if allow_unbounded else 'No'} range allowed to be unbounded;" \
                f" this is the {ordinal(len(ranges)+1)} range in {paths}"
            unbounded		= True
            ranges[c]		= lambda b=b: itertools.count( b )
        path_segs[i]		= f"{{{c}}}" + ( "'" if tic else "" )

    path_fmt			= '/'.join( path_segs )
    return path_fmt, ranges


def path_sequence(
    path_fmt: str,
    ranges: Dict[str, Callable[[], int]],
):
    """Yield a sequence of paths, modulating the format specifiers of the
    path_fmt according to their value sources in ranges.

    For example, a

        path_fmt = "m/44'/60'/0'/0/{f}", with a
        ranges   = dict( f=lambda b=0, e=2: range( b, e+1 ) )

    would yield the paths:

        "m/44'/60'/0'/0/0"
        "m/44'/60'/0'/0/1"
        "m/44'/60'/0'/0/2"
    """
    # Stert all the iterators
    viters			= {
        k: iter( l() )
        for k,l in ranges.items()
    }
    values			= {		# Initial round of values; must provide at least one
        k: next( viters[k] )
        for k in viters
    }
    assert all( v is not None for v in values.values() ), \
        "Iterators for path segment values must yield at least an initial value"

    while not any( v is None for v in values.values() ):
        yield path_fmt.format( **values )
        if not ranges:
            break				# No variable records at all; just one
        # Get the next value.  Working from the lowest iterator up, cycle value(s)
        for i,k in enumerate( sorted( viters.keys(), reverse=True )):
            values[k]		= next( viters[k], None )
            if values[k] is not None:
                break
            # OK, this iterable has ended.  Restart it, and cycle to the next one up, iff
            # there are remaining ranges
            if i+1 < len( ranges ):
                viters[k]	= iter( ranges[k]() )
                values[k]	= next( viters[k], None )


def cryptopaths_parser( cryptocurrency, edit=None ):
    """
    Generate a standard cryptopaths list, from the given list of "<crypto>[:<paths>]"
    cryptocurrencies (default: CRYPTO_PATHS).

    Adjusts the provided derivation paths by an optional eg. "../-" path adjustment."""
    cryptopaths 		= []
    for crypto in cryptocurrency or CRYPTO_PATHS:
        try:
            crypto,paths	= crypto.split( ':' )
        except ValueError:
            crypto,paths	= crypto,None
        crypto			= Account.supported( crypto )
        if paths is None:
            paths		= Account.path_default( crypto )
        if edit:
            paths		= path_edit( paths, edit )
        cryptopaths.append( (crypto,paths) )
    return cryptopaths


def random_secret(
    seed_length			= BITS_DEFAULT // 8
) -> bytes:
    """Generates a new random secret.

    NOTE: There is a slightly less than 1 / 2^128 chance that any given random secret will lead to
    an invalid BTC wallet private key!  This is because the 256-bit seed for bitcoin must be less than
    the secp256k1 field size:

        0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141

    We cannot generate secrets that are guaranteed to be valid for every derived HD Wallet BTC
    address.  The secret is derived through a complex hashing procedure for each wallet at each
    path.  The probability of this occurring is so vanishingly small that we will instead simply opt
    to *not* generate an public wallet address, when asked.

    Just to give you an understandable reference, the first 128 bits of an illegal secret unusable
    for generating a Bitcoin wallet must have 128 '1' bits at its start.  If wallets were generated
    at a billion per second, by every person alive on earth, it should take about 1.5 trillion years
    to arrive at the first invalid seed.

    2**128 / 1e9 / 7e9 / (60*60*24*365) == 1,541,469,010,115.145

    """
    return RANDOM_BYTES( seed_length )


Details = namedtuple( 'Details', ('name', 'group_threshold', 'groups', 'accounts') )


def enumerate_mnemonic( mnemonic ):
    if isinstance( mnemonic, str ):
        mnemonic		= mnemonic.split( ' ' )
    return dict(
        (i, f"{i+1:>2d} {w}")
        for i,w in enumerate( mnemonic )
    )


def organize_mnemonic( mnemonic, rows=None, cols=None, label="" ):
    """Given a SLIP-39 "word word ... word" or ["word", "word", ..., "word"] mnemonic, emit rows
    organized in the desired rows and cols (with defaults, if not provided).  We return the fully
    formatted line, plus the list of individual words in that line.

    """
    num_words			= enumerate_mnemonic( mnemonic )
    if not rows or not cols:
        rows,cols		= MNEM_ROWS_COLS.get( len(num_words), (7, 3))
    for r in range( rows ):
        line			= label if r == 0 else ' ' * len( label )
        words			= []
        for c in range( cols ):
            word		= num_words.get( c * rows + r )
            if word:
                words.append( word )
                line	       += f"{word:<13}"
        yield line,words


def group_parser( group_spec ):
    """Parse a SLIP-39 group specification.

        Frens6, Frens 6, Frens(6)	- A 3/6 group (default is 1/2 of group size, rounded up)
        Frens2/6, Frens(2/6)		- A 2/6 group

    """
    match			= group_parser.RE.match( group_spec )
    if not match:
        raise ValueError( f"Invalid group specification: {group_spec!r}" )
    name			= match.group( 'name' )
    size			= match.group( 'size' )
    require			= match.group( 'require' )
    if not size:
        size			= 1
    if not require:
        # eg. default 2/4, 3/5
        require			= math.ceil( int( size ) * GROUP_REQUIRED_RATIO )
    return name,(int(require),int(size))
group_parser.RE			= re.compile( # noqa E305
    r"""^
        \s*
        (?P<name> [^\d\(/]+ )
        \s*\(?\s*
        (:? (?P<require> \d* ) \s* / )?
        \s*
        (?P<size> \d* )
        \s*\)?\s*
        $""", re.VERBOSE )


def create(
    name: str,
    group_threshold: int,
    groups: Dict[str,Tuple[int, int]],
    master_secret: bytes	= None,		# Default: 128-bit seeds
    passphrase: bytes		= b"",
    iteration_exponent: int	= 1,
    cryptopaths: Sequence[Tuple[str,str]] = None,  # default: ETH, BTC at default paths
    strength: int		= 128,
) -> Tuple[str,int,Dict[str,Tuple[int,List[str]]], Sequence[Sequence[Account]]]:
    """Creates a SLIP-39 encoding and 1 or more Ethereum accounts.  Returns the details, in a form
    compatible with the output API.

    """
    if master_secret is None:
        assert strength in BITS, f"Invalid {strength}-bit secret length specified"
        master_secret		= random_secret( strength // 8 )
    g_names,g_dims		= list( zip( *groups.items() ))
    mnems			= mnemonics(
        group_threshold	= group_threshold,
        groups		= g_dims,
        master_secret	= master_secret,
        passphrase	= passphrase,
        iteration_exponent= iteration_exponent
    )
    # Derive the desired account(s) at the specified derivation paths, or the default
    accts			= list( accountgroups(
        master_secret	= master_secret,
        cryptopaths	= cryptopaths or [('ETH',None), ('BTC',None)],
        allow_unbounded	= False,
    ))

    groups			= {
        g_name: (g_of, g_mnems)
        for (g_name,(g_of, _),g_mnems) in zip( g_names, g_dims, mnems )
    }
    if log.isEnabledFor( logging.INFO ):
        group_reqs			= list(
            f"{g_nam}({g_of}/{len(g_mns)})" if g_of != len(g_mns) else f"{g_nam}({g_of})"
            for g_nam,(g_of,g_mns) in groups.items() )
        requires		= f"Recover w/ {group_threshold} of {len(groups)} groups {', '.join(group_reqs)}"
        for g_n,(g_name,(g_of,g_mnems)) in enumerate( groups.items() ):
            log.info( f"{g_name}({g_of}/{len(g_mnems)}): {requires}" )
            for mn_n,mnem in enumerate( g_mnems ):
                for line,_ in organize_mnemonic( mnem, label=f"{ordinal(mn_n+1)} " ):
                    log.info( f"{line}" )

    return Details(name, group_threshold, groups, accts)


def mnemonics(
    group_threshold: int,
    groups: Sequence[Tuple[int, int]],
    master_secret: Union[str,bytes] = None,
    passphrase: bytes		= b"",
    iteration_exponent: int	= 1,
    strength: int		= 128,
) -> List[List[str]]:
    """Generate SLIP39 mnemonics for the supplied group_threshold of the given groups.  Will generate a
     random master_secret, if necessary.

    """
    if master_secret is None:
        assert strength in BITS, f"Invalid {strength}-bit secret length specified"
        master_secret		= random_secret( strength // 8 )
    if len( master_secret ) * 8 not in BITS:
        raise ValueError(
            f"Only {', '.join( f'{b}-' for b in BITS )}bit seeds supported; {len(master_secret)*8}-bit master_secret supplied" )
    return generate_mnemonics(
        group_threshold	= group_threshold,
        groups		= groups,
        master_secret	= master_secret,
        passphrase	= passphrase,
        iteration_exponent = iteration_exponent )


def account(
    master_secret: Union[str,bytes],
    crypto: str			= None,  # default 'ETH'
    path: str			= None,  # default to the crypto's path_default
    format: str			= None,  # eg. 'bech32', or use the default address format for the crypto
):
    """Generate an HD wallet Account from the supplied master_secret seed, at the given HD derivation
    path, for the specified cryptocurrency.

    """
    acct			= Account(
        crypto		= crypto or 'ETH',
        format		= format,
    ).from_seed(
        seed		= master_secret,
        path		= path,
    )
    return acct


def accounts(
    master_secret: Union[str,bytes],
    crypto: str			= None,  # default 'ETH'
    paths: str			= None,  # default to the crypto's path_default; allow ranges
    format: str			= None,
    allow_unbounded		= True,
):
    """Create accounts for crypto, at the provided paths (allowing ranges), with the optionsal address format. """
    for path in [None] if paths is None else path_sequence( *path_parser(
        paths		= paths,
        allow_unbounded	= allow_unbounded,
    )):
        yield account( master_secret, crypto=crypto, path=path, format=format )


def accountgroups(
    master_secret: bytes,
    cryptopaths: Sequence[Tuple[str,str]],
    allow_unbounded: bool	= True,
) -> Sequence[Sequence[Account]]:
    """Generate the desired cryptocurrency account(s) at each crypto's given path(s).  This is useful
    for generating sequences of groups of wallets for multiple cryptocurrencies, eg. for receiving
    multiple cryptocurrencies for each client.  Since each cryptocurrency uses a different BIP-44 path,
    we have to generate different sequences.

    Supports ranges in each path segment, eg.:

        ('ETH', "m/44'/60'/0'/0/-")	-- generates all accounts for ETH
        ('BTC', "m/44'/0'/0'/0/-")	-- generates all accounts for BTC

        [
          [ "m/44'/60'/0'/0/0", "0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1"],
          [ "m/44'/0'/0'/0/0", "1MAjc529bjmkC1iCXTw2XMHL2zof5StqdQ"]
        ],
        [
          [ "m/44'/60'/0'/0/1", "0x8D342083549C635C0494d3c77567860ee7456963"],
          [ "m/44'/0'/0'/0/1", "1BGwDuVPJeXDG9upaHvVPds5MXwkTjZoav"]
        ],
        ...

    """
    yield from zip( *[
        accounts(
            master_secret	= master_secret,
            paths		= paths,
            crypto		= crypto,
            allow_unbounded	= allow_unbounded,
        )
        for crypto,paths in cryptopaths
    ])


def address(
    master_secret: bytes,
    crypto: str			= None,
    path: str			= None,
    format: str			= None,
):
    """Return the specified cryptocurrency HD account address at path."""
    return account(
        master_secret,
        path		= path,
        crypto		= crypto,
        format		= format,
    ).address


def addresses(
    master_secret: bytes,
    crypto: str	 		= None,  # default 'ETH'
    paths: str			= None,  # default: The crypto's path_default; supports ranges
    format: str			= None,
    allow_unbounded: bool	= True,
):
    """Generate a sequence of cryptocurrency account (path, address, ...)  for all designated
    cryptocurrencies.  Usually a single (<path>, <address>) tuple is desired (different
    cryptocurrencies typically have their own unique path derivations.

    """
    for acct in accounts( master_secret, crypto, paths, format, allow_unbounded=allow_unbounded ):
        yield (acct.crypto, acct.path, acct.address)


def addressgroups(
    master_secret: bytes,
    cryptopaths: Sequence[Tuple[str,str]],
    allow_unbounded: bool	= True,
) -> Sequence[str]:
    """Yields account (<crypto>, <path>, <address>) records for the desired cryptocurrencies at paths.

    """
    yield from zip( *[
        addresses(
            master_secret	= master_secret,
            paths		= paths,
            crypto		= crypto,
            allow_unbounded	= allow_unbounded,
        )
        for crypto,paths in cryptopaths
    ])



import ast
import io
import os
import json
import logging
import math

from datetime		import datetime
from collections	import namedtuple
from collections.abc	import Callable

from typing		import Dict, List, Tuple, Sequence, Any
import qrcode
from fpdf		import FPDF, FlexTemplate

from ..api		import Account, cryptopaths_parser, create, enumerate_mnemonic, group_parser
from ..util		import input_secure, ordinal, chunker
from ..defaults		import (
    FONTS, MNEM_ROWS_COLS, CARD, CARD_SIZES, PAPER, PAGE_MARGIN, MM_IN, PT_IN,
    WALLET, WALLET_SIZES, COLOR,
    GROUPS, GROUP_THRESHOLD_RATIO,
    FILENAME_FORMAT,
)

# Optionally support output of encrypted JSON files
eth_account			= None
try:
    import eth_account
except ImportError:
    pass

log				= logging.getLogger( __package__ )


class Region:
    """Takes authority for a portion of another Region, and places things in it, relative to its
    upper-left and lower-right corners.

    Attempts to create a sane priority hierarchy, because fpdf2's FlexTemplate sorts based on it.
    If a specific priority is provided, it will be used.  However, if no priority is given, 0 will
    be automatically assigned by FlexTemplate.  If sorting is unstable, this may result in an
    unspecified ordering.

    """
    def __init__( self, name, x1=None, y1=None, x2=None, y2=None, rotate=None, priority=None ):
        self.name		= name
        self.x1			= x1		# could represent positions, offsets or ratios
        self.y1			= y1
        self.x2			= x2
        self.y2			= y2
        self.rotate		= rotate
        self.priority		= priority
        self.regions		= []

    @property
    def h( self ):
        return self.y2 - self.y1

    @property
    def w( self ):
        return self.x2 - self.x1

    def element( self ):
        "Converts to mm.  Optionally returns a specified priority."
        d			= dict(
            name	= self.name,
            x1		= self.x1 * MM_IN,
            y1		= self.y1 * MM_IN,
            x2		= self.x2 * MM_IN,
            y2		= self.y2 * MM_IN,
            rotate	= self.rotate or 0.0
        )
        if self.priority:
            d['priority']	= self.priority
        return d

    def __str__( self ):
        return f"({self.name:16}: ({float(self.x1):8}, {float( self.y1):8}) - ({float(self.x2):8} - {float(self.y2):8})"

    def add_region( self, region ):
        region.x1		= self.x1 if region.x1 is None else region.x1
        region.y1		= self.y1 if region.y1 is None else region.y1
        region.x2		= self.x2 if region.x2 is None else region.x2
        region.y2		= self.y2 if region.y2 is None else region.y2

        self.regions.append( region )
        return region

    def add_region_relative( self, region ):
        region.x1	       += self.x1
        region.y1	       += self.y1
        region.x2	       += self.x2
        region.y2	       += self.y2
        self.regions.append( region )
        return region

    def add_region_proportional( self, region ):
        region.x1		= self.x1 + self.w * ( 0 if region.x1 is None else region.x1 )
        region.y1		= self.y1 + self.h * ( 0 if region.y1 is None else region.y1 )
        region.x2		= self.x1 + self.w * ( 1 if region.x2 is None else region.x2 )
        region.y2		= self.y1 + self.h * ( 1 if region.y2 is None else region.y2 )
        self.regions.append( region )
        return region

    def dimensions( self ):
        "Converts to mm."
        return Coordinate( self.w * MM_IN, self.h * MM_IN )

    def elements( self ):
        """Yield a sequence of { 'name': "...", 'x1': #,  ... }."""
        if self.__class__ != Region:
            yield self.element()
        for r in self.regions:
            for d in r.elements():
                yield d


class Text( Region ):
    SIZE_RATIO			= 3/4

    def __init__( self, *args, font=None, text=None, size=None, size_ratio=None, align=None,
                  foreground=None, background=None, bold=None, italic=None, underline=None,
                  **kwds ):
        self.font		= font
        self.text		= text
        self.size		= size
        self.size_ratio		= size_ratio or self.SIZE_RATIO
        self.align		= align
        self.foreground		= foreground
        self.background		= background
        self.bold		= bold
        self.italic		= italic
        self.underline		= underline
        super().__init__( *args, **kwds )

    def element( self ):
        d			= super().element()
        d['type']		= 'T'
        d['font']		= FONTS.get( self.font ) or self.font or FONTS.get( 'sans' )
        line_height		= self.h * PT_IN  # Postscript point == 1/72 inch
        d['size']		= self.size or ( line_height * self.size_ratio )  # No need for int(round(..))
        if self.text is not None:
            d['text']		= self.text
        if self.bold:
            d['bold']		= True
        if self.italic:
            d['italic']		= True
        if self.underline:
            d['underline']	= True
        if self.align is not None:
            d['align']		= self.align
        if self.foreground is not None:
            d['foreground']	= self.foreground
        if self.background is not None:
            d['background']	= self.background
        return d


class Image( Region ):
    def __init__( self, *args, font=None, text=None, size=None, **kwds ):
        self.font		= font
        self.text		= text
        self.size		= None
        super().__init__( *args, **kwds )

    def element( self ):
        d			= super().element()
        d['type']		= 'I'
        return d


class Line( Region ):
    def emit( self, d, *args, **kwds ):
        d['type']		= 'L'
        super().emit( *args, **kwds )


class Box( Region ):
    def element( self ):
        d			= super().element()
        d['type']		= 'B'
        return d


Coordinate			= namedtuple( 'Coordinate', ('x', 'y') )


def layout_card(
    card_size: Coordinate,
    card_margin: int,
    num_mnemonics: int	= 20,
):
    card			= Box( 'card', 0, 0, card_size.x, card_size.y )
    card_interior		= card.add_region_relative(
        Region( 'card-interior', x1=+card_margin, y1=+card_margin, x2=-card_margin, y2=-card_margin )
    )
    for c_n in range( 16 ):  # SLIP-39 supports up to 16 groups
        card_interior.add_region_proportional(
            Text( f'card-g{c_n}', x1=1/8, y1=-1/16, x2=7/8, y2=5/16, foreground=int( COLOR[c_n % len( COLOR )], 16 ), rotate=-45 )
        )
    card_top			= card_interior.add_region_proportional(
        Region( 'card-top', x1=0, y1=0, x2=1, y2=1/4 )
    )
    card_bottom			= card_interior.add_region_proportional(
        Region( 'card-bottom', x1=0, y1=1/4, x2=1, y2=1 )
    )
    card_mnemonics		= card_bottom.add_region_proportional(
        Region( 'card-mnemonics', x1=0, y1=0, x2=13/16, y2=1 )
    )

    # QR codes sqaare, and anchored to top and bottom of card.
    card_qr1			= card_bottom.add_region_proportional(
        Image( 'card-qr1', x1=13/16, y1=0, x2=1, y2=1/2 )
    )
    card_qr1_size		= min( card_qr1.w, card_qr1.h )
    card_qr1.x1			= card_qr1.x2 - card_qr1_size
    card_qr1.y2			= card_qr1.y1 + card_qr1_size

    card_qr2			= card_bottom.add_region_proportional(
        Image( 'card-qr2', x1=13/16, y1=1/2, x2=1, y2=1 )
    )
    card_qr2_size		= min( card_qr2.w, card_qr2.h )
    card_qr2.x1			= card_qr2.x2 - card_qr2_size
    card_qr2.y1			= card_qr2.y2 - card_qr2_size

    card_top.add_region_proportional(
        Text( 'card-title', x1=0, y1=0, x2=1, y2=40/100, bold=True )
    )
    card_top.add_region_proportional(
        Text( 'card-requires', x1=0, y1=40/100, x2=1, y2=66/100, align='C', italic=True )
    )
    card_top.add_region_proportional(
        Text( 'card-crypto1', x1=0, y1=66/100, x2=1, y2=83/100, align='R' )
    )
    card_top.add_region_proportional(
        Text( 'card-crypto2', x1=0, y1=83/100, x2=1, y2=100/100, align='R' )
    )

    assert num_mnemonics in MNEM_ROWS_COLS, \
        f"Invalid SLIP-39 mnemonic word count: {num_mnemonics}"
    rows,cols		= MNEM_ROWS_COLS[num_mnemonics]
    for r in range( rows ):
        for c in range( cols ):
            card_mnemonics.add_region_proportional(
                Text(
                    f"mnem-{c * rows + r}",
                    x1		= c/cols,
                    y1		= r/rows,
                    x2		= (c+1)/cols,
                    y2		= (r+1)/rows,
                    font	= 'mono',
                    size_ratio	= 9/16,
                )
            )

    return card


def layout_wallet(
    wallet_size: Coordinate,
    wallet_margin: int,
):
    """Produce a template format for a Paper Wallet.

    Sets priority:
      -3     : Hindmost backgrounds
      -2     : Things atop background, but beneath contrast-enhancement
      -1     : Contrast-enhancing partially transparent images
       0     : Text, etc. (the default)
    """
    prio_backing		= -3
    prio_normal			= -2
    prio_contrast		= -1

    wallet			= Box( 'wallet', 0, 0, wallet_size.x, wallet_size.y )
    wallet_background		= wallet.add_region_proportional(  # noqa: F841
        Image( 'wallet-bg', priority=prio_backing ),
    )

    # Most of the Paper Wallet is visible (contains public address information).  The right-most
    # portion can be folded over, and displays private key information
    show			= wallet.add_region_proportional(
        Region( 'wallet-show', x2=5/8 )
    )
    fold			= wallet.add_region_proportional(
        Box( 'wallet-fold', x1=5/8 )
    )

    public			= show.add_region_relative(
        Region( 'wallet-public', x1=+wallet_margin, y1=+wallet_margin, x2=-wallet_margin, y2=-wallet_margin )
    )
    private			= fold.add_region_relative(
        Region( 'wallet-private', x1=+wallet_margin, y1=+wallet_margin, x2=-wallet_margin, y2=-wallet_margin )
    )

    # Assign each different Crypto name a different color, in template labels crypto-{f,b}1, crypto-{f,b}2, ...
    for c_n in range( len( COLOR )):
        public.add_region_proportional(
            Text( f'crypto-f{c_n}', x1=1/8, y1=-1/16, x2=7/8, y2=7/16, foreground=int( COLOR[c_n], 16 ), rotate=-45, priority=prio_normal )
        )
        private.add_region_proportional(
            Text( f'crypto-b{c_n}', x1=2/8, y1=-1/16, x2=7/8, y2=7/16, foreground=int( COLOR[c_n], 16 ), rotate=-45, priority=prio_normal )
        )

    # The background rosette and cryptocurrency symbol in the center
    public_center_size		= min( public.w, public.h )
    public.add_region(
        Image(
            'center',
            x1		= public.x1 + public.w * 2/3 - public_center_size / 3,
            y1		= public.y1 + public.h * 1/2 - public_center_size / 3,
            x2		= public.x1 + public.w * 2/3 + public_center_size / 3,
            y2		= public.y1 + public.h * 1/2 + public_center_size / 3,
        )
    )

    # Public addresses are vertical on left- and right-hand of public Region.  In order to fit the
    # longer ETH addresses into a space with a fixed-width font, we know that the ratio of the width
    # to the height has to be about 1/20.  Rotation is downward around upper-left corner; so,
    # lower-left corner will shift 1 height leftward and upward; so start 1 height right and down.
    address_length		= public.h
    address_height		= address_length * 1/20
    public.add_region(
        Image(
            'address-l-bg',
            x1		= public.x1 + address_height,
            y1		= public.y1,
            x2		= public.x1 + address_height + address_length,
            y2		= public.y1 + address_height,
            rotate	= -90,
            priority	= prio_contrast,
        )
    ).add_region(
        Text(
            'address-l',
            font	= 'mono',
            rotate	= -90,
        )
    )

    # Rotation is upward around the upper-left corner, so lower-left corner will shift 1 height
    # upward and right; so start 1 height leftward and down.
    public.add_region(
        Image(
            'address-r-bg',
            x1		= public.x2 - address_height,
            y1		= public.y2,
            x2		= public.x2 - address_height + address_length,
            y2		= public.y2 + address_height,
            rotate	= +90,
            priority	= prio_contrast,
        )
    ).add_region(
        Text(
            'address-r',
            font	= 'mono',
            rotate	= +90,
        )
    )

    # Wallet name, amount
    public.add_region_proportional(
        Text( 'name-label',	x1=1/16, y1=0/16, x2=4/16, y2=1/16 )
    )
    public.add_region_proportional(
        Image( 'name-bg',	x1=4/16, y1=0/16, x2=15/16, y2=1/16, priority=prio_contrast )
    )
    public.add_region_proportional(
        Text( 'name',		x1=4/16, y1=0/16, x2=15/16, y2=1/16 )
    )

    public.add_region_proportional(
        Text( 'amount-label',	x1=1/16, y1=3/16, x2=4/16, y2=1/16 )
    )
    public.add_region_proportional(
        Image( 'amount-bg',	x1=4/16, y1=0/16, x2=15/16, y2=1/16, priority=prio_contrast )
    )
    public.add_region_proportional(
        Text( 'amount',		x1=4/16, y1=0/16, x2=15/16, y2=1/16 )
    )

    # Make Public QR Code square w/ min of height, width, anchored at lower-left corner
    public.add_region_proportional(
        Text( 'address-qr-t',	x1=1/16, y1=5/16, x2=1, y2=6/16 )
    )
    public_qr			= public.add_region_proportional(
        Image( 'address-qr-bg',	x1=1/16, y1=6/16, x2=1, y2=15/16 )
    )
    public_qr_size		= min( public_qr.w, public_qr.h )
    public_qr.x2		= public_qr.x1 + public_qr_size
    public_qr.y1		= public_qr.y2 - public_qr_size
    public_qr.add_region(
        Image( 'address-qr' )
    )

    public.add_region_proportional(
        Text( 'address-qr-b',	x1=1/16, y1=15/16, x2=1, y2=16/16 )
    )
    public_qr_r			= public.add_region_proportional(
        Text( 'address-qr-r',	x1=1/16, y1=6/16, x2=1, y2=7/16, rotate=-90 )
    )
    public_qr_r.x1	       += public_qr_size + public_qr_r.h
    public_qr_r.x2	       += public_qr_size

    # Private region

    private.add_region_proportional(
        Text( 'private-qr-t',	x1=0/16, y1=5/16, x2=1, y2=6/16 )
    )
    private_qr			= private.add_region_proportional(
        Image( 'private-qr-bg',	x1=0/16, y1=6/16, x2=1, y2=15/16, priority=prio_contrast )
    )
    # QR code at most 5/8 the width, to retain sufficient space for large Ethereum encrypted JSON
    # wallet private keys
    private_qr_size		= min( private_qr.w, private_qr.h, private.w * 5 / 8 )
    private_qr.x2		= private_qr.x1 + private_qr_size
    private_qr.y1		= private_qr.y2 - private_qr_size
    private_qr.add_region(
        Image( 'private-qr' )
    )
    private.add_region_proportional(
        Text( 'private-qr-b',	x1=0/16, y1=15/16, x2=1, y2=16/16 )
    )

    # Hint above; but same computed width as private_qr
    private_h_t			= private.add_region_proportional(
        Text( 'private-hint-t',	x1=0/16, y1=0/16, x2=1, y2=1/16 )
    )
    private_h_t.x2		= private_qr.x2
    private_h_bg		= private.add_region_proportional(
        Image( 'private-hint-bg',x1=0/16, y1=1/16, x2=1, y2=4/16, priority=prio_contrast )
    )
    private_h_bg.x2		= private_qr.x2
    private_h			= private.add_region_proportional(
        Text( 'private-hint',	x1=0/16, y1=1/16, x2=1, y2=3/16 )
    )
    private_h.x2		= private_qr.x2

    # We'll use the right side of the private region, each line rotated 90 degrees down and right.
    # So, we need the upper-left corner of the private-bg anchored at the upper-right corner of
    # private.
    private_length		= private.h                     # line length is y-height
    private_fontsize		= 6.5				# points == 1/72 inch
    private_height		= private.x2 - private_qr.x2 - .05
    private_lineheight		= private_fontsize / PT_IN / Text.SIZE_RATIO * .9  # in.

    private.add_region(
        Image(
            'private-bg',
            x1		= private.x2,
            y1		= private.y1,
            x2		= private.x2 + private_length,
            y2		= private.y1 + private_height,
            rotate	= -90,
            priority	= prio_contrast,
        )
    )
    # Now, add private key lines down the edge from right to left, rotating each into place
    for ln in range( int( private_height // private_lineheight )):
        private.add_region(
            Text(
                f"private-{ln}",
                font	= 'mono',
                x1	= private.x2 - private_lineheight * ln,
                y1	= private.y1,
                x2	= private.x2 + private_length,
                y2	= private.y1 + private_lineheight,
                size	= private_fontsize,
                rotate	= -90,
            )
        )

    return wallet


def layout_components(
    pdf: FPDF,
    comp_dim: Coordinate,
    page_margin_mm: float	= .25 * MM_IN,   # 1/4" in mm.
) -> Tuple[int, Callable[[int],[int, Coordinate]]]:
    """Compute the number of components per pdf page, and a function returning the page # and
    component (x,y) for the num'th component."""

    # FPDF().epw/.eph is *without* page margins, but *with* re-orienting for portrait/landscape
    # Allow 5% bleed over into page margins (to allow for slight error in paper vs. comp sizes)
    comp_cols			= int(( pdf.epw - page_margin_mm * 2 * 95/100 ) // comp_dim.x )
    comp_rows			= int(( pdf.eph - page_margin_mm * 2 * 95/100 ) // comp_dim.y )
    comps_pp			= comp_rows * comp_cols

    def page_xy( num ):
        """Returns the page, and the coordinates within that page of the num'th component"""
        page,nth		= divmod( num, comps_pp )
        page_rows,page_cols	= divmod( nth, comp_cols )
        offsetx			= page_margin_mm + page_cols * comp_dim.x
        offsety			= page_margin_mm + page_rows * comp_dim.y
        log.debug( f"{ordinal(num)} {comp_dim.x:7.5f}mm x {comp_dim.y:7.5f}mm component on page {page}, offset {offsetx:7.5f}mm x {offsety:7.5f}mm" )
        return (page, Coordinate( x=offsetx, y=offsety ))

    return (comps_pp, page_xy)


def layout_pdf(
        card_dim: Coordinate,                   # mm.
        page_margin_mm: float	= .25 * MM_IN,  # 1/4" in mm.
        orientation: str	= 'portrait',
        paper_format: Any	= PAPER         # Can be a paper name (Letter) or (x, y) dimensions in mm.
) -> Tuple[int, str, Callable[[int],[int, Coordinate]], FPDF]:
    """Find the ideal orientation for the most cards of the given dimensions.  Returns the number of
    cards per page, the FPDF, and a function useful for laying out templates on the pages of the
    PDF."""
    pdf				= FPDF(
        orientation	= orientation,
        format		= paper_format,
    )
    pdf.set_margin( 0 )

    cards_pp,page_xy		= layout_components(
        pdf, comp_dim=card_dim, page_margin_mm=page_margin_mm
    )
    return cards_pp, orientation, page_xy, pdf


def output_pdf(
    name: str,
    group_threshold: int,
    groups: Dict[str,Tuple[int,List[str]]],
    accounts: Sequence[Sequence[Account]],
    card_format: str		= CARD,   # 'index' or '(<h>,<w>),<margin>'
    paper_format: Any		= PAPER,  # 'Letter', (x,y) dimensions in mm.
) -> Tuple[FPDF, Sequence[Sequence[Account]]]:
    """Produces a PDF, containing a number of cards containing the provided slip39.Details if
    card_format is not False, and pass through the supplied accounts."""

    # Deduce the card size
    try:
        (card_h,card_w),card_margin = CARD_SIZES[card_format.lower()]
    except KeyError:
        (card_h,card_w),card_margin = ast.literal_eval( card_format )
    card_size			= Coordinate( y=card_h, x=card_w )

    # Compute how many cards per page.  Flip page portrait/landscape to match the cards'.  Use the length
    # of the first Group's first Mnemonic to determine the number of mnemonics on each card.
    num_mnemonics		= len( groups[next(iter(groups.keys()))][1][0].split() )
    card			= layout_card(  # converts to mm
        card_size, card_margin, num_mnemonics=num_mnemonics )
    card_dim			= card.dimensions()

    # Find the best PDF and orientation, by max of returned cards_pp (cards per page).  Assumes
    # layout_pdf returns a tuple that can be compared; cards_pp,orientation,... will always sort.
    page_margin_mm		= PAGE_MARGIN * MM_IN
    cards_pp,orientation,page_xy,pdf = max(
        layout_pdf( card_dim, page_margin_mm, orientation=orientation, paper_format=paper_format )
        for orientation in ('portrait', 'landscape')
    )
    log.debug( f"Page: {paper_format} {orientation} {pdf.epw:.8}mm w x {pdf.eph:.8}mm h w/ {page_margin_mm}mm margins,"
               f" Card: {card_format} {card_dim.x:.8}mm w x {card_dim.y:.8}mm h == {cards_pp} cards/page" )

    elements			= list( card.elements() )
    if log.isEnabledFor( logging.DEBUG ):
        log.debug( f"Card elements: {json.dumps( elements, indent=4)}" )
    tpl				= FlexTemplate( pdf, list( card.elements() ))

    group_reqs			= list(
        f"{g_nam}({g_of}/{len(g_mns)})" if g_of != len(g_mns) else f"{g_nam}({g_of})"
        for g_nam,(g_of,g_mns) in groups.items() )
    requires			= f"Recover w/ {group_threshold} of {len(group_reqs)} groups {', '.join(group_reqs[:4])}{'...' if len(group_reqs)>4 else ''}"

    # Convert all of the first group's account(s) to an address QR code
    assert accounts and accounts[0], \
        "At least one cryptocurrency account must be supplied"
    qr				= {}
    for i,acct in enumerate( accounts[0] ):
        qrc			= qrcode.QRCode(
            version	= None,
            error_correction = qrcode.constants.ERROR_CORRECT_M,
            box_size	= 10,
            border	= 1
        )
        qrc.add_data( acct.address )
        qrc.make( fit=True )

        qr[i]			= qrc.make_image()
        if log.isEnabledFor( logging.INFO ):
            f			= io.StringIO()
            qrc.print_ascii( out=f )
            f.seek( 0 )
            for line in f:
                log.info( line.strip() )

    card_n			= 0
    page_n			= None
    for g_n,(g_name,(g_of,g_mnems)) in enumerate( groups.items() ):
        for mn_n,mnem in enumerate( g_mnems ):
            p,(offsetx,offsety)	= page_xy( card_n )
            if p != page_n:
                pdf.add_page()
                page_n		= p
            card_n	       += 1

            tpl['card-title']	= f"SLIP39 {g_name}({mn_n+1}/{len(g_mnems)}) for: {name}"
            tpl['card-requires'] = requires
            tpl['card-crypto1']	= f"{accounts[0][0].crypto} {accounts[0][0].path}: {accounts[0][0].address}"
            tpl['card-qr1']	= qr[0].get_image()
            if len( accounts[0] ) > 1:
                tpl['card-crypto2'] = f"{accounts[0][1].crypto} {accounts[0][1].path}: {accounts[0][1].address}"
                tpl['card-qr2']	= qr[1].get_image()
            tpl[f'card-g{g_n}']	= f"{g_name:6.6}..{mn_n+1}" if len(g_name) > 7 else f"{g_name} {mn_n+1}"

            for n,m in enumerate_mnemonic( mnem ).items():
                tpl[f"mnem-{n}"] = m

            tpl.render( offsetx=offsetx, offsety=offsety )

    return pdf,accounts


def write_pdfs(
    names		= None,		# sequence of [ <name>, ... ], or { "<name>": <details> }
    master_secret	= None,		# Derive SLIP-39 details from this seed (if not supplied in names)
    passphrase		= None,		# UTF-8 encoded passphrase; default is '' (empty)
    group		= None,		# Group specifications, [ "Frens(3/6)", ... ]
    group_threshold	= None,		# int, or 1/2 of groups by default
    cryptocurrency	= None,		# sequence of [ 'ETH:<path>', ... ] to produce accounts for
    edit		= None,		# Adjust crypto paths according to the provided path edit
    card_format		= None,		# Eg. "credit"; False outputs no SLIP-39 Mnemonic cards to PDF
    paper_format	= None,		# Eg. "Letter", "Legal", (x,y)
    filename		= None,		# A file name format w/ {path}, etc. formatters
    filepath		= None,		# and optionally a path
    json_pwd		= None,		# If JSON wallet output desired, supply password
    text		= None,		# Truthy outputs SLIP-39 recover phrases to stdout
    wallet_pwd		= None,		# If paper wallet images desired, supply password
    wallet_pwd_hint	= None,		# an optional password hint
    wallet_format	= None,		# and a paper wallet format (eg. 'quarter')
    wallet_paper	= None,		# default Wallets to output on Letter format paper
):
    """Writes a PDF containing a unique SLIP-39 encoded seed for each of the names specified.

    Returns a { "<filename>": <details>, ... } dictionary of all PDF files written, and each of
    their account details.

    """
    # Convert sequence of group specifications into standard { "<group>": (<needs>, <size>) ... }
    if isinstance( group, dict ):
        groups			= group
    else:
        groups			= dict(
            group_parser( g )
            for g in group or GROUPS
        )
    assert groups and all( isinstance( k, str ) and len( v ) == 2 for k,v in groups.items() ), \
        f"Each group member specification must be a '<group>': (<needs>, <size>), not {type(next(groups.items()))}"

    group_threshold		= int( group_threshold ) if group_threshold else math.ceil( len( groups ) * GROUP_THRESHOLD_RATIO )

    cryptopaths			= cryptopaths_parser( cryptocurrency, edit=edit )

    # If account details not provided in names, generate them.
    if not isinstance( names, dict ):
        assert not master_secret or not names or len( names ) == 1, \
            "Creating multiple account details from the same secret entropy doesn't make sense"
        names			= {
            name: create(
                name		= name,
                group_threshold	= group_threshold,
                groups		= groups,
                master_secret	= master_secret,
                passphrase	= passphrase.encode( 'utf-8' ) if passphrase else b'',
                cryptopaths	= cryptopaths,
            )
            for name in names or [ "SLIP39" ]
        }

    # Generate each desired SLIP-39 Mnemonic file.  Supports --card (the default).
    results			= {}
    for name, details in names.items():
        # Get the first group of the accountgroups in details.accounts.
        accounts		= details.accounts
        assert accounts and accounts[0], \
            "At least one --cryptocurrency must be specified"
        for account in accounts[0]:
            log.warning( f"{account.crypto:6} {account.path:20}: {account.address}" )

        if text:
            # Output the SLIP-39 mnemonics as text:
            #    name: mnemonic
            for g_name,(g_of,g_mnems) in details.groups.items():
                for mnem in g_mnems:
                    print( f"{name}{name and ': ' or ''}{mnem}" )

        # Unless no card or paper wallets specified, output a PDF containing the SLIP-39 mnemonic recovery cards
        pdf			= None
        if card_format is not False or wallet_pwd:
            pdf,_		= output_pdf(
                *details,
                card_format	= card_format or CARD,
                paper_format	= paper_format or PAPER )

        now			= datetime.now()

        pdf_name		= ( filename or FILENAME_FORMAT ).format(
            name	= name,
            date	= datetime.strftime( now, '%Y-%m-%d' ),
            time	= datetime.strftime( now, '%H.%M.%S'),
            crypto	= accounts[0][0].crypto,
            path	= accounts[0][0].path,
            address	= accounts[0][0].address,
        )
        if not pdf_name.lower().endswith( '.pdf' ):
            pdf_name	       += '.pdf'
        if filepath:
            pdf_name		= os.path.join( filepath, pdf_name )

        if wallet_pwd:
            # Deduce the paper wallet size and create a template.  All layouts are in specified in
            # inches; template dimensions are in mm.
            try:
                (wall_h,wall_w),wall_margin = WALLET_SIZES[wallet_format.lower() if wallet_format else WALLET]
            except KeyError:
                (wall_h,wall_w),wall_margin = ast.literal_eval( wallet_format )

            wall		= layout_wallet( Coordinate( y=wall_h, x=wall_w ), wall_margin )  # converts to mm
            wall_dim		= wall.dimensions()

            # Lay out wallets, always in Portrait orientation, defaulting to the Card paper_format
            # if it is a standard size (a str, not an (x,y) tuple), otherwise to "Letter" paper.
            if wallet_paper is None:
                wallet_paper	= paper_format if type(paper_format) is str else PAPER

            pdf.add_page( orientation='P', format=wallet_paper )
            page_margin_mm	= PAGE_MARGIN * MM_IN

            walls_pp,page_xy	= layout_components( pdf, comp_dim=wall_dim, page_margin_mm=page_margin_mm )
            elements		= list( wall.elements() )
            if log.isEnabledFor( logging.DEBUG ):
                log.debug( f"Wallet elements: {json.dumps( elements, indent=4)}" )
            wall_tpl		= FlexTemplate( pdf, elements )

            # Place each Paper Wallet adding pages as necessary (we already have the first fresh page).
            wall_n		= 0
            page_n		= 0
            for account_group in accounts:
                for c_n,account in enumerate( account_group ):
                    p,(offsetx,offsety) = page_xy( wall_n )
                    if p != page_n:
                        pdf.add_page( orientation='P', format=wallet_paper )
                        page_n	= p
                    try:
                        private_enc		= account.encrypted( 'password' )
                    except NotImplementedError as exc:
                        log.exception( f"{account.crypto} doesn't support BIP-38 or JSON wallet encryption: {exc}" )
                        continue

                    wall_n     += 1

                    images			= os.path.dirname( __file__ )
                    wall_tpl['wallet-bg']	= os.path.join( images, 'paper-wallet-background.png' )
                    wall_tpl[f"crypto-f{c_n}"]	= account.crypto
                    wall_tpl[f"crypto-b{c_n}"]	= account.crypto

                    wall_tpl['center']		= os.path.join( images, account.crypto + '.png' )

                    wall_tpl['name-label']	= "Wallet:"
                    wall_tpl['name-bg']		= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['name']		= name

                    # wall_tpl['center-bg']	= os.path.join( images, 'guilloche-center.png' )

                    public_qr	= qrcode.QRCode(
                        version		= None,
                        error_correction = qrcode.constants.ERROR_CORRECT_M,
                        box_size	= 10,
                        border		= 1,
                    )
                    public_qr.add_data( account.address )
                    wall_tpl['address-l-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['address-l']	= account.address
                    wall_tpl['address-r-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['address-r']	= account.address

                    wall_tpl['address-qr-t']	= 'PUBLIC ADDRESS'
                    wall_tpl['address-qr-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['address-qr']	= public_qr.make_image( back_color="transparent" ).get_image()
                    wall_tpl['address-qr-b']	= 'DEPOSIT/VERIFY'
                    wall_tpl['address-qr-r']	= account.path

                    private_qr	= qrcode.QRCode(
                        version		= None,
                        error_correction = qrcode.constants.ERROR_CORRECT_M,
                        box_size	= 10,
                        border		= 1,
                    )
                    private_qr.add_data( private_enc )

                    wall_tpl['private-bg']	= os.path.join( images, '1x1-ffffff54.png' )

                    # If not enough lines, will throw Exception, as it should!  We don't want to
                    # emit a Paper Wallet without the entire encrypted private key present.  This is
                    # primarily an issue for Ethereum encrypted JSON wallets, which are very large.
                    # As the vertical aspect ratio of the Paper Wallet increases (eg. for half- or
                    # third-page vs. quarter-page Paper Wallets), the line length increases, but the
                    # number of lines available decreases.  Estimate the number of characters on
                    # each line.
                    line_elm			= next( e for e in elements if e['name'] == 'private-0' )
                    line_points			= ( line_elm['x2'] - line_elm['x1'] ) / MM_IN * PT_IN
                    line_fontsize		= line_elm['size']
                    line_chars			= int( line_points / line_fontsize / ( 5 / 8 ))  # Chars ~ 5/8 width vs. height
                    log.debug( f"Private key line length: {line_chars} chars" )
                    for ln,line in enumerate( chunker( private_enc, line_chars )):
                        wall_tpl[f"private-{ln}"] = line
                    wall_tpl['private-hint-t']	= 'PASSWORD HINT:'
                    wall_tpl['private-hint-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['private-hint']	= wallet_pwd_hint
                    wall_tpl['private-qr-t']	= 'PRIVATE KEY'
                    wall_tpl['private-qr-bg']	= os.path.join( images, '1x1-ffffff54.png' )
                    wall_tpl['private-qr']	= private_qr.make_image( back_color="transparent" ).get_image()
                    wall_tpl['private-qr-b']	= 'SPEND'

                    wall_tpl.render( offsetx=offsetx, offsety=offsety )

        if json_pwd:
            # If -j|--json supplied, also emit the encrypted JSON wallet.  This may be a *different*
            # password than the SLIP-39 master secret encryption passphrase.  It will be required to
            # decrypt and use the saved JSON wallet file, eg. to load a software Ethereum wallet.
            assert eth_account, \
                "The optional eth-account package is required to support output of encrypted JSON wallets\n" \
                "    python3 -m pip install eth-account"
            assert any( 'ETH' == crypto for crypto,paths in cryptopaths ), \
                "--json is only valid if '--crypto ETH' wallets are specified"
            if json_pwd == '-':
                json_pwd	= input_secure( 'JSON key file password: ', secret=True )
            else:
                log.warning( "It is recommended to not use '-j|--json <password>'; specify '-' to read from input" )

            for eth in (
                account
                for group in accounts
                for account in group
                if account.crypto == 'ETH'
            ):
                json_str	= json.dumps( eth_account.Account.encrypt( eth.key, json_pwd ), indent=4 )
                json_name	= ( filename or FILENAME_FORMAT ).format(
                    name	= name or "SLIP39",
                    date	= datetime.strftime( now, '%Y-%m-%d' ),
                    time	= datetime.strftime( now, '%H.%M.%S'),
                    crypto	= eth.crypto,
                    path	= eth.path,
                    address	= eth.address,
                )
                if json_name.lower().endswith( '.pdf' ):
                    json_name	= json_name[:-4]
                json_name      += '.json'
                while os.path.exists( json_name ):
                    log.error( "ERROR: Will NOT overwrite {json_name}; adding '.new'!" )
                    json_name  += '.new'
                with open( json_name, 'w' ) as json_f:
                    json_f.write( json_str )
                log.warning( f"Wrote JSON {name or 'SLIP39'}'s encrypted ETH wallet {eth.address} derived at {eth.path} to: {json_name}" )

                if pdf:
                    # Add the encrypted JSON account recovery to the PDF also, if generated.
                    pdf.add_page()
                    margin_mm	= PAGE_MARGIN * MM_IN
                    pdf.set_margin( 1.0 * MM_IN )

                    col_width	= pdf.epw - 2 * margin_mm
                    pdf.set_font( FONTS['sans'], size=10 )
                    line_height	= pdf.font_size * 1.2
                    pdf.cell( col_width, line_height, json_name )
                    pdf.ln( line_height )

                    pdf.set_font( FONTS['sans'], size=9 )
                    line_height	= pdf.font_size * 1.1

                    for line in json_str.split( '\n' ):
                        pdf.cell( col_width, line_height, line )
                        pdf.ln( line_height )
                    pdf.image( qrcode.make( json_str ).get_image(), h=min(pdf.eph, pdf.epw)/2, w=min(pdf.eph, pdf.epw)/2 )

        if pdf:
            pdf.output( pdf_name )
            log.warning( f"Wrote SLIP39-encoded wallet for {name!r} to: {pdf_name}" )
        results[pdf_name]	= details

    return results
